require 'mrlib'

module Model3DS
   class Vertex < MetaRed::Vector
      def to_s
         "#{@x} #{@y} #{@z}"
      end
   end
   class Color < MetaRed::Vector
      alias :r :x
      alias :g :y
      alias :b :z
      def to_s
         "#{@x} #{@y} #{@z}"
      end
   end
   class Mesh
      attr_accessor :materials, :objects, :master_scale, :ambient_light
      def initialize
         @materials = []
         @objects = []
      end
      def self.load(dat, chunk_end)
         mesh = Mesh.new
         while(dat.pos < chunk_end)
            chunk_id, chunk_size = dat.get_chunk
            case chunk_id
            when 0x0100
               mesh.master_scale = dat.get_float
            when 0x2100
               mesh.ambient_light = dat.get_color
            when 0xAFFF
               mesh.materials << Material.load(dat, dat.pos + chunk_size - 6)
            when 0x4000
               obj_name = dat.get_string
               mesh.objects << Object.load(obj_name, dat, dat.pos + chunk_size - 6)
            else
               dat.skip
            end
         end
         dat.seek(chunk_end)
         return mesh
      end
      def save_as_mqo(file)
         open(file, 'w'){|out|
            out.print "Metasequoia Document\nFormat Text Ver 1.0\n"
            out.print "Material #{@materials.size} {\n"
            mat_id_tbl = Hash.new
            @materials.each_with_index{|x,i|
               x.fputs_as_mqo(out)
               mat_id_tbl[x.name] = i
            }
            out.print "}\n"
            @objects.each{|x|
               x.fputs_as_mqo(out, mat_id_tbl)
            }
            out.print "Eof\n"
         }
      end
   end
   class Material
      attr_accessor(
         :name, :ambient, :diffuse, :specular, :shininess, :shin2pct, :shin3pct, :transparency,
         :xpfall, :refblur, :self_ilpct, :shading, :mapping
      )
      def initialize
         @mappings = []
      end
      def self.load(dat, chunk_end)
         mat = Material.new
         while(dat.pos < chunk_end)
            chunk_id, chunk_size = dat.get_chunk
            case chunk_id
            when 0xA000
               mat.name = dat.get_string
            when 0xA010
               mat.ambient = dat.get_color
            when 0xA020
               mat.diffuse = dat.get_color
            when 0xA030
               mat.specular = dat.get_color
            when 0xA040
               mat.shininess = dat.get_percentage
            when 0xA041
               mat.shin2pct = dat.get_percentage
            when 0xA042
               mat.shin3pct = dat.get_percentage
            when 0xA050
               mat.transparency = dat.get_percentage
            when 0xA052
               mat.xpfall = dat.get_percentage
            when 0xA053
               mat.refblur = dat.get_percentage
            when 0xA084
               mat.self_ilpct = dat.get_percentage
            when 0xA100
               mat.shading = dat.get_short
            when 0xA200
               mat.mapping = Mapping.load(dat, dat.pos + chunk_size - 6)
            else
               dat.skip
            end
         end
         dat.seek(chunk_end)
         return mat
      end
      def fputs_as_mqo(out)
         out.print "	\"#{@name}\" col(#{@diffuse*(1.0/256)} #{(256.0-@transparency)/256}) #{@mapping}\n"
      end
   end
   class Mapping
      attr_accessor :power, :file, :uscale, :vscale, :uoffset, :voffset
      def self.load(dat, chunk_end)
         map = Mapping.new
         map.power = dat.get_percentage

         while(dat.pos < chunk_end)
            chunk_id, chunk_size = dat.get_chunk
            case chunk_id
            when 0xA300
               map.file = dat.get_string
            when 0xA354
               map.uscale = dat.get_float
            when 0xA356
               map.vscale = dat.get_float
            when 0xA358
               map.uoffset = dat.get_float
            when 0xA35A
               map.voffset = dat.get_float
            else
               dat.skip
            end
         end
         dat.seek(chunk_end)
         return map
      end
      def to_s
         "tex(\"#{@file}\")"
      end
   end
   class Object
      attr_accessor :name, :vertices, :faces, :uv_array, :groups, :matrix
      def initialize(name)
         @name = name
         @groups = Hash.new
      end
      def self.load(obj_name, dat, chunk_end)
         obj = Object.new(obj_name)
         while(dat.pos < chunk_end)
            chunk_id, chunk_size = dat.get_chunk
            case chunk_id
            when 0x4100
               #print " tri_obj\n" if $Debug3ds
            when 0x4110
               obj.vertices = dat.get_point_array
            when 0x4120
               obj.faces = dat.get_face_array
            when 0x4130
               mat_name = dat.get_string
               nfaces = dat.get_short
               obj.groups[mat_name] = dat.read(2*nfaces).unpack('S*')
            when 0x4140
               nUVs = dat.get_short
               obj.uv_array = Array.new(nUVs)
               obj.uv_array.map!{|x|
                  dat.read(8).unpack('f2')
               }
            when 0x4160
               obj.matrix = Array.new(4)
               obj.matrix.map!{|x|
                  dat.read(12).unpack('f3')
               }
            else
               dat.skip
            end
         end
         dat.seek(chunk_end)
         return obj
      end
      def fputs_as_mqo(out, mat_id_tbl)
         out.print "Object \"#{@name}\" {\n"
         out.print "	vertex #{@vertices.size} {\n"
         @vertices.each{|x|
            out.print "\t\t#{x}\n"
         }
         out.print "	}\n"
         out.print "	face #{@faces.size} {\n"
         @groups.each{|mat, grp|
            mat_id = mat_id_tbl[mat]
            grp.each{|i|
               tmp = @faces[i][0]
               next if tmp.uniq.size != 3
               out.print "		3 V(#{tmp[0]} #{tmp[2]} #{tmp[1]}) M(#{mat_id})"
               unless @uv_array.nil?
                  out.print " UV("
                  out.print [0,2,1].map{|i| "#{@uv_array[tmp[i]][0]} #{1-@uv_array[tmp[i]][1]}"}.join(' ')
                  out.print ")"
               end
               out.print "\n"
            }
         }
         out.print "	}\n"
         out.print "}\n"
      end
   end

   def self.load(file)
      ret = nil
      open(file, 'rb'){|dat|
         instance_eval{
            def dat.get_chunk
               begin
                  @chunk_id = self.read(2).unpack('S')[0]
                  @chunk_size = self.read(4).unpack('L')[0]
                  STDOUT.print "0x#{(self.pos-6).to_s(16)}:[0x#{@chunk_id.to_s(16)}, #{@chunk_size}byte(0x#{@chunk_size.to_s(16)})]\n" if $Debug3ds
                  return @chunk_id, @chunk_size
               rescue NoMethodError
                  return :eof
               end
            end
            def dat.skip
               self.seek(@chunk_size-6, IO::SEEK_CUR)
            end
            def dat.get_float
               self.read(4).unpack('f')[0]
            end
            def dat.get_short
               self.read(2).unpack('S')[0]
            end
            def dat.get_string
               self.gets("\0").strip
            end
            def dat.get_point_array
               npoints = self.get_short
               parray = Array.new(npoints)
               parray.map!{|x|
                  Vertex.new(* self.read(12).unpack('f3'))
               }
               return parray
            end
            def dat.get_face_array
               nfaces = self.get_short
               farray = Array.new(nfaces)
               farray.map!{|x|
                  # [ vertex_id_ary, flag ]
                  [self.read(6).unpack('S3'), self.get_short]
               }
               return farray
            end
            def dat.get_color
               get_chunk
               case @chunk_id
               when 0x0010, 0x0013
                  return Color.new(* self.read(12).unpack('f3'))
               when 0x0011, 0x0012
                  return Color.new(* self.read(3).unpack('C3'))
               else
                  self.skip
                  return nil
               end
            end
            def dat.get_percentage
               get_chunk
               case @chunk_id
               when 0x0030
                  return self.read(2).unpack('S')[0]
               when 0x0031
                  return self.read(4).unpack('f')[0]
               else
                  self.skip
                  return nil
               end
            end
         }
         chunk_id, chunk_size = dat.get_chunk
         if chunk_id == 0x4D4D
            loop{
               chunk_id, chunk_size = dat.get_chunk
               case chunk_id
               when 0x3D3D
                  ret = Mesh.load(dat, dat.pos + chunk_size - 6)
               when :eof
                  break
               else
                  dat.skip
               end
            }
         end
      }
      return ret
   end
end
